/*
 *  PowerPC software driven TLB support for qemu.
 *  This code should be common to all targets, then I put it in a separate
 *  code file.
 * 
 *  Copyright (c) 2005 Jocelyn Mayer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

static void soft_tlb_invalidate_one (CPUState *env, ppc_tlb_t *tlb,
                                     int qinval)
{
    if (!(tlb->prot & PAGE_VALID))
        return;
#if defined (DEBUG_SOFTWARE_TLB)
    if (!likely(loglevel == 0)) {
        fprintf(logfile, "Invalidate TLB %08lx %08lx %08lx %02x\n",
                (unsigned long)tlb->VPN, (unsigned long)tlb->RPN,
                (unsigned long)tlb->size, tlb->prot);
    }
#endif
    if (likely(qinval)) {
#if !defined(FLUSH_ALL_TLBS)
        target_phys_addr_t page;

        for (page = tlb->VPN;
             page < tlb->VPN + tlb->size; page += TARGET_PAGE_SIZE) {
            tlb_flush_page(env, page);
        }
#else
        tlb_flush(env, 1);
#endif
    }
    tlb->prot &= ~PAGE_VALID;
}

void soft_tlb_invalidate_all (CPUState *env)
{
    int i;

#if defined (DEBUG_SOFTWARE_TLB)
    if (!likely(loglevel == 0)) {
        fprintf(logfile, "Invalidate all TLBs\n");
    }
#endif
    /* Invalidate all defined software TLB */
    for (i = 0; i < env->nb_tlb; i++)
        soft_tlb_invalidate_one(env, &env->tlb[i], 0);
    tlb_flush(env, 1);
}

void soft_tlb_invalidate_virt (CPUState *env, target_ulong address)
{
    ppc_tlb_t *tlb;
    int i;

    for (i = 0; i < env->nb_tlb; i++) {
        tlb = &env->tlb[i];
        if ((tlb->prot & PAGE_VALID) &&
            address >= tlb->VPN && address < (tlb->VPN + tlb->size))
            soft_tlb_invalidate_one(env, tlb, 1);
    }
}

void soft_tlb_store (CPUState *env, int nr, target_phys_addr_t RPN,
                     target_ulong VPN, target_ulong size,
                     int prot, int is_user, uint32_t flags, uint32_t private)
{
    ppc_tlb_t *tlb;

    tlb = &env->tlb[nr];
    /* If TLB is valid, invalidate it */
    if (tlb->prot & PAGE_VALID)
        soft_tlb_invalidate_one(env, tlb, 1);
#if defined (DEBUG_SOFTWARE_TLB)
    if (!likely(loglevel == 0)) {
        fprintf(logfile, "Set TLB %d on %08lx %08lx %08lx prot=%c%c%c "
                "user=%d %c\n",
                nr, (unsigned long)VPN, (unsigned long)RPN, (unsigned long)size,
                prot & PAGE_READ ? 'R' : '-', prot & PAGE_WRITE ? 'W' : '-',
                prot & PAGE_EXEC ? 'X' : '-', is_user,
                prot & PAGE_VALID ? 'V' : '-');
    }
#endif
    tlb->RPN = RPN & ~(size - 1);
    tlb->VPN = VPN & ~(size - 1);
    tlb->size = size;
    tlb->prot = prot;
    tlb->is_user = is_user;
    tlb->flags = flags;
    tlb->private = private;
}

int soft_tlb_check_one (CPUState *env, target_phys_addr_t *physical, int *prot,
                        int nr, target_ulong vaddr, int rw,
                        int access_type, int is_user)
{
    ppc_tlb_t *tlb;
    int ret, tp;

    tlb = &env->tlb[nr];
    tp = -1;
    ret = -1;
#if defined (DEBUG_SOFTWARE_TLB)
    if (!likely(loglevel == 0) && (tlb->prot & PAGE_VALID)) {
        fprintf(logfile, "TLB %d %d check %08x <= %08x < %08x %d %d a %d %d p %d\n",
                nr, env->nb_tlb, tlb->VPN, vaddr, tlb->VPN + tlb->size,
                is_user, tlb->is_user, access_type, rw, tlb->prot);
    } else if (!likely(loglevel == 0)) {
        fprintf(logfile, "TLB %d %d NV %08x <= %08x < %08x %d %d a %d %d p %d\n",
                nr, env->nb_tlb, tlb->VPN, vaddr, tlb->VPN + tlb->size,
                is_user, tlb->is_user, access_type, rw, tlb->prot);
    }
#endif
    if ((tlb->prot & PAGE_VALID) && tlb->is_user == is_user &&
        vaddr >= tlb->VPN && vaddr < (tlb->VPN + tlb->size)) {
        if ((access_type == ACCESS_CODE && !(tlb->prot & PAGE_EXEC)) ||
            (rw == 1 && !(tlb->prot & PAGE_WRITE)) ||
            (rw == 0 && !(tlb->prot & PAGE_READ))) {
            /* Access denied */
            ret = -2;
        } else if (env->tlb_check_more == NULL ||
                   (ret = (*env->tlb_check_more)(env, tlb, &tp, vaddr, rw,
                                                 access_type, is_user)) >= 0) {
            if (tp == -1)
                tp = tlb->prot;
            *physical = tlb->RPN | (vaddr & (tlb->size - 1));
            *prot = tp;
#if defined (DEBUG_SOFTWARE_TLB)
            if (!likely(loglevel == 0)) {
                fprintf(logfile, "Found TLB %d for %08lx: %08lx %08lx "
                        "prot=%c%c%c user=%d rw %d %d %08x\n",
                        nr, (unsigned long)vaddr, (unsigned long)tlb->RPN,
                        (unsigned long)tlb->size,
                        *prot & PAGE_READ ? 'R' : '-',
                        *prot & PAGE_WRITE ? 'W' : '-',
                        *prot & PAGE_EXEC ? 'X' : '-', tlb->is_user,
                        rw, is_user, *physical);
            }
#endif
            return 0;
        }
    }

    return ret;
}

int soft_tlb_find (CPUState *env, target_phys_addr_t *physical, int *prot,
                   target_ulong vaddr, int rw, int access_type, int is_user)
{
    int i, tmp, ret;

    ret = -1;
    for (i = 0; i < env->nb_tlb; i++) {
        tmp = soft_tlb_check_one(env, physical, prot, i, vaddr, rw,
                                 access_type, is_user);
        if (ret >= 0)
            return 0;
        if (ret == -1 && tmp != -1)
            ret = tmp;
    }

    return ret;
}
